#!/usr/bin/perl

#TODO
# - Improvements depend on improvements to SOAP::Lite
#   Can we somehow use doc/lit support of SOAP::Lite?? 

use warnings;
use strict;

use Data::Dumper;

use Getopt::Long 2.32 qw(:config auto_help auto_version bundling);
use Pod::Usage;

# The following modules are only for testing/diffing, but they should be
# available on all Perl installations.
use IO::File;
use POSIX qw(tmpnam);

use FindBin;
use lib "$FindBin::RealBin/../lib";

my $ns_nmwg = "http://ggf.org/ns/nmwg/base/2.0/";


my $soap_prefix = <<__PREFIX__;
<?xml version="1.0" encoding="UTF-8"?>
<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
                  xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
<soapenv:Body>
__PREFIX__
my $soap_suffix = '</soapenv:Body>\n</soapenv:Envelope>';

my ($tmpname,$tmpfh); # variables for tmp files used in END{}

# Parse command line parameters

my $module = "soap";
my $host = "";
my $port = "";
my $endpoint = "";
my $uri = "";
my $reqfile = undef;
my $test = undef;

GetOptions(
  "module|m=s"      => \$module,
  "host|h=s"        => \$host,
  "port|p=i"        => \$port,
  "endpoint|e=s"    => \$endpoint,
  "uri|u=s"         => \$uri,
  "reqfile|f=s"     => \$reqfile,
  "test|t=s"          => \$test,
) or pod2usage(2);

my @reqfiles = @ARGV;
push @reqfiles,$reqfile if $reqfile;
pod2usage("Mandatory parameter missing.\n")
  unless @reqfiles && ($host || $uri);

unless ($uri) {
  # Build URI from other parameters
  $uri = "http://$host"; # $host || $uri - see above
  $uri .= ":$port" if $port;
  if ($endpoint) {
    $endpoint = "/$endpoint" unless $endpoint =~ m#^/#;
    $uri .= $endpoint;
  } else {
    $uri .= "/";
  }
}

my $diff;
if ($test && $test ne "set") {
  if ($test && ($test eq "XmlDiff" || $test eq "default")) {
    require XML::Diff;
    $diff = XML::Diff->new();
  }
  open LOG, ">perfsonar-client-test.log";
}

foreach my $reqfile (@reqfiles) {
  # Read (and parse) request file
  open REQ, $reqfile or die "Could not open $reqfile\n";
  my $req = "";
  foreach my $line (<REQ>) {
    $req .= $line unless $line =~ m/^<\?xml.*/;
  }
  close REQ;


  # Now do the hard work
  my $result;
  if ($module eq "soap") {
    require "NMWG.pm";
    require "NMWG/Message.pm";
    require "perfSONAR.pm";
    require "perfSONAR/SOAP.pm";
    require "perfSONAR/SOAP/Message.pm";
    require "perfSONAR/SOAP/HTTP/Request.pm";
    require "perfSONAR/SOAP/HTTP/Response.pm";
    require "perfSONAR/SOAP/HTTP/UserAgent.pm";
    
    my $request = perfSONAR::SOAP::Message->new(body => $req, uri => $uri);
    my $userAgent = perfSONAR::SOAP::HTTP::UserAgent->new;
    my $http_request = perfSONAR::SOAP::HTTP::Request->new(message => $request);
    my $http_response = $userAgent->request($http_request);
    unless ($http_response->is_success) {
      my $code = $http_response->code();
      my $message = $http_response->message();
      print "HTTP Error: $code, $message\n";
      exit;
    }
    my $soap_response = $http_response->soap_message;
    if ($soap_response->is_fault) {
      die "SOAP Error!\n" .
        "  faultstring: ". ($soap_response->{fault}->faultstring || "") .
        "  faultcode: ". join(" - ", $soap_response->{fault}->faultcode) ."\n";
    }
    $result = $soap_response->as_string;
      
  } elsif ($module eq "sendReceive") {
    require "NMWG.pm"; 
    require "NMWG/Message.pm"; 
    require "perfSONAR.pm";
    require "perfSONAR/SOAP.pm";
    require "perfSONAR/SOAP/Message.pm";
    require "perfSONAR/SOAP/HTTP/Request.pm";
    require "perfSONAR/SOAP/HTTP/Response.pm";
    require "perfSONAR/SOAP/HTTP/UserAgent.pm";
    my $request = NMWG::Message->new($req);
    my $response = perfSONAR::sendReceive(
      message => $request,
      uri => $uri
    );

    $result = $response->as_string(2);

  } elsif ($module eq "http") {
    require LWP::UserAgent;
    require HTTP::Request;
    require HTTP::Response;

    my $httpEndpoint = $uri || "http://$host:$port$endpoint";

    my $userAgent = LWP::UserAgent->new(timeout => 5000);

    my $sendSoap = HTTP::Request->new(
      POST => $httpEndpoint,
      new HTTP::Headers,
      $soap_prefix . $req . $soap_suffix
    );
    $sendSoap->header('SOAPAction' => '""');
    $sendSoap->content_type  ('text/xml');
    $sendSoap->content_length(length($req));
    $sendSoap->protocol("HTTP/1.0");
    $sendSoap->header('pragma' => 'no-cache');
    $sendSoap->header('cache-control' => 'no-cache');
    $sendSoap->header('accept' => 'application/soap+xml, application/dime, multipart/related, text/*');
    my $httpResponse = $userAgent->request($sendSoap);
    my $responseCode = $httpResponse->code();
    my $responseContent = $httpResponse->content();
    $result = $httpResponse->as_string(2);

  } elsif ($module eq "raw"){

    require LWP::UserAgent;
    require HTTP::Request;
    require HTTP::Response;

    my $httpEndpoint = $uri || "http://$host:$port$endpoint";

    my $userAgent = LWP::UserAgent->new(timeout => 5000);

    my $sendSoap = HTTP::Request->new(
      POST => $httpEndpoint,
      new HTTP::Headers,
      $req 
    );
    $sendSoap->header('SOAPAction' => '""');
    $sendSoap->content_type  ('text/xml');
    $sendSoap->content_length(length($req));
    $sendSoap->protocol("HTTP/1.0");
    $sendSoap->header('pragma' => 'no-cache');
    $sendSoap->header('cache-control' => 'no-cache');
    $sendSoap->header('accept' => 'application/soap+xml, application/dime, multipart/related, text/*');
    my $httpResponse = $userAgent->request($sendSoap);
    my $responseCode = $httpResponse->code();
    my $responseContent = $httpResponse->content();
    $result = $httpResponse->as_string(2);

    
  } else {
    pod2usage("Unkown module type: $module.\n");
  }

  if ($test) {
    my $testfile = "test/$reqfile";
    if ($test eq "set") {
      open TESTFILE, ">$testfile"
        or die "Could not create test file $testfile\n";
      print TESTFILE $result;
      close TESTFILE;
    } elsif ($test eq "XmlDiff" || $test eq "default") {
      print "Testing $reqfile ... ";
      unless (open TESTFILE, "<$testfile") {
        print LOG ">>> $testfile NOT FOUND\n";
        print "NOT FOUND\n";
        next;
      }
      my $save = $/;
      $/ = undef;
      my $old_xml = <TESTFILE>;
      $/ = $save;
      close TESTFILE;
      my $diffgram;
      eval {
        $diffgram = $diff->compare(-old => $old_xml, -new => $result);
      };
      if ($@) {
        print LOG ">>> Error parsing XML:\n$@";
        print "ERROR\n";
        next;
      }
      if ($diffgram->documentElement()->hasChildNodes()) {
        print LOG ">>> Diff for $reqfile:\n" . $diffgram->toString(2);
        print "CHANGED\n";
        next;
      }
      print "ok\n";
    } elsif ($test eq "xmldiff" || $test eq "diff" || $test eq "3dm" ||
             $test eq "diffxml" || $test eq "XmlSemanticDiff") {
      print "Testing $reqfile ... ";
      my ($out,$err);
      if ($test eq "xmldiff" || $test eq "3dm" || $test eq "diffxml" ||
          $test eq "XmlSemanticDiff" ) { 
        # try new temporary filenames until we get one that didn't already exist
        do { $tmpname = tmpnam() }
        until $tmpfh = IO::File->new($tmpname, O_RDWR|O_CREAT|O_EXCL);
        $tmpfh->print($result);
        $tmpfh->close();
      } elsif ($test eq "diff") {
        require IPC::Run;
        IPC::Run::run(['diff',"--version"],\undef,\$out,\$err,
                      IPC::Run::timeout(10))
          or die "Cannot execute diff: $?\n";
      }
      if (! -e $testfile) {
        $err = "File not found: $testfile\n";
      } elsif (! -r $testfile) {
        $err = "File not readable: $testfile\n";
      } else {
        if ($test eq "xmldiff") {
          # Seems like 'xmldiff' does not use stderr!
          # Return codes are also strange.
          require IPC::Run;
          IPC::Run::run(['xmldiff',$testfile,$tmpname],
            \undef, \$out, \$err, IPC::Run::timeout(10));
          if ($?>>8 == 255) {
            $err = $out;
            $out = "";
          }
        } elsif ($test eq "diff") {
          # If $testfile doesn't exist, "run" reports only a "broken pipe"
          # => We really need the check for the test file from above
          require IPC::Run;
          IPC::Run::run(['diff',"-u","-d","-w",$testfile,"-"],
            \$result, \$out, \$err, IPC::Run::timeout(10));
          if ($?) {
            my $result = $? >> 8;
            my $signal = $? & 127;
            if ($result >= 2) {
              $err = $out;
              $out = "";
            } elsif ($signal) {
              $err = "Caught signal: $signal\n";
            }
          } else {
            # No differences
            $out = "";
          }
        } elsif ($test eq "3dm") {
          # Seems to just throw exceptions on failure...
          # Return code always "0" ...
          require IPC::Run;
          IPC::Run::run(['3dm',"-d",$testfile,$tmpname],
            \undef, \$out, \$err, IPC::Run::timeout(10));
          if ($out =~ m#<diff />$#) {
            $out = "";
          }
        } elsif ($test eq "diffxml") {
          # TODO for "-x":
          #   Exception in thread "main" java.lang.NoClassDefFoundError:
          #     org/apache/xpath/XPathAPI
          # Return code seems to be "1" for "changes" and Exception. Error
          # lead to "2". But it's easier to distinguish using stdout<->stderr
          require IPC::Run;
          IPC::Run::run(['diffxml',"-f",$testfile,$tmpname],
            \undef, \$out, \$err, IPC::Run::timeout(10));
          if ($?>>8 == 0) { # or: ($out =~ m#<delta/>$#)
            $out = "";
          }
        } elsif ($test eq "XmlSemanticDiff") {
          require XML::SemanticDiff;
          my $diff = XML::SemanticDiff->new(keeplinenums => 1);
          foreach my $change ($diff->compare($testfile,$tmpname)) {
            $out .= "lines $change->{startline}-$change->{endline}: " .
              $change->{message};
          }
        }
      }
      if ($err) {
        print LOG ">>> Error calling $test:\n$err";
        print "ERROR\n";
        next;
      }
      if ($out) {
        print LOG ">>> Diff for $reqfile:\n" . $out;
        print "CHANGED\n";
        next;
      }
      print "ok\n";
    } else {
      pod2usage("Unknown test module: $test.\n");
    }
  } else {
    print "Response for $reqfile:\n\n";
    print $result;
    print "\n\n\n";
  }
}

close LOG;

exit 0;


## END OF MAIN ##


END {
  if ($tmpname) {
    undef $tmpfh;  # automatically closes the file if it's open
    -e $tmpname && unlink($tmpname) or die "Couldn't unlink $tmpname : $!";
  }
}


__END__



=head1 NAME

B<perfsonar-client.pl> - Send request to a perfSONAR service

=head1 SYNOPSIS

B<perfsonar-client.pl> S<[B<--help>]> S<[B<--module>|B<-m>=nmwg|soap|http]>
  ( S<B<--host>|B<-h>=HOSTNAME> S<B<--port>|B<-p>=PORT> S<B<--endpoint>|B<-p>=ENDPOINT> ) |
  ( S<B<--uri>|B<-u>=URI> )
  S<[B<--test>=set|xmldiff|...]>
  S<B<--reqfile>|B<-f>=F<REQFILE>> | REQFILE1 REQFILE2 ...



=head1 DESCRIPTION

Sends a request to a perfSONAR service and displays response.



=head1 OPTIONS


=over


=item B<--help>

Prints a help message and exits.


=item B<--module>|B<-m>=soap|http|nmwg

Use Perl module SOAP::Lite (soap) or Perl module LWP::UserAgent (http) for
accessing the service.

IMPORTANT: SOAP::Lite only does rpc/encoded and LWP::UserAgent only does
document/literal SOAP requests!


=item B<--host>|B<-h>=HOST

Connect to service on host HOST. You have to at least set B<--host> or
B<--uri>.


=item B<--port>|B<-p>=PORT

Use port PORT to connect. Default is to use default HTTP port (80).


=item B<--endpoint>|B<-u>=ENDPOINT

Use ENDPOINT as endpoint of the service. 


=item B<--uri>|B<-u>=URI

Use URI as service uri. You have to at least set B<--host> or
B<--uri>. If B<--uri> is not set it is build based on http://HOST:PORT/ENDPOINT.
If B<--uri> is set, B<--host>, B<--port>, and B<--endpoint> are ignored!


=item B<--reqfile>|B<-f>=F<REQFILE>

Use content of F<REQFILE> as request.

=item B<--test>|B<-t>=MODE

Simple test of responses by "diffing" with result from previous run.
Available values for MODE:
  set       Save result for testing.
            No diffing done!
  XmlDiff   Use XML::Diff.
            Often very verbose...
  XmlSemanticDiff  Use XML::SemanticDiff.
            Looks very promising!
  xmldiff   Use xmldiff (http://www.logilab.org/projects/xmldiff/).
            Very famous Python tool. Seems to produce live locks
            and quite long deltas.
  3dm       Use 3dm (http://www.cs.hut.fi/~ctl/3dm/).
            Fast and small, but "unreadable" deltas. Java.
  diffxml   Use diffxml (http://diffxml.sourceforge.net/)
            Java. Two algorithms:
            FMES algorithm (seems to produce live locks)
            xmdiff (produces exceptions).
  diff      Yes, "diff". Nothing more, but also nothing less.
Binaries/scripts have to be in your PATH, of course!


=head1 EXAMPLES


S<perfsonar-client.pl> S<--host=localhost> S<--port=5000> \
  S<--uri=http://test/Hades/perfSONAR/MA> S<--reqfile=SetupData-Req.xml>

S<perfsonar-client.pl> S<-h selena.acad.bg> S<-p 8090> \
  S<-f SetupDataRequest-FilterRRDSelect-ACAD.xml>
